## @namespace officegenerator.libxlsxgenerator
## @brief Este módulo permite la lectura y escritura de ficheros xlsx de Microsoft Excel
##
## You can change current sheet, with createSheet or using setCurrentSheet. After that all commands will use that sheet until you change it again
import gettext
import openpyxl
import openpyxl.comments
import openpyxl.cell
import openpyxl.styles
import openpyxl.worksheet
import openpyxl.worksheet.worksheet
import openpyxl.formatting.rule
import pkg_resources

from decimal import Decimal
from officegenerator.commons import columnAdd, Coord, Range, topLeftCellNone, convert_command, generate_formula_total_string, Coord_from_index
from officegenerator.objects.currency import Currency, currency_symbol
from officegenerator.objects.formula import isFormula
from officegenerator.objects.percentage import Percentage
from os import path, makedirs,  sep
from tempfile import TemporaryDirectory

try:
    t=gettext.translation('officegenerator',pkg_resources.resource_filename("officegenerator","locale"))
    _=t.gettext
except:
    _=str

class ColumnWidthXLSX:
    Date=40
    Detetime=60

class XLSX_Commons:
    def __init__(self):
        pass

    ## It returns a sheet object with the index id
    def get_sheet_by_id(self, id):
        return self.wb[self.wb.sheetnames[id]]

    ## It returns a index integer of the sheet with a given name
    def get_sheet_id(self, name):
        for id, s_name in enumerate(self.wb.sheetnames):
            if s_name==name:
                return id
        return None

    ## Function that establishes current worksheet. Updates self.ws_current and self.ws_current_id
    ##
    ## id Is a integer beginning with 0
    ## name is the title of the sheet
    ## @param id_or_name Index or Nmae
    def setCurrentSheet(self, id_or_name):
        if id_or_name.__class__==int:
            self.ws_current_id=id_or_name
        else:#name
            self.ws_current_id=self.get_sheet_id(id_or_name)
        self.ws_current=self.get_sheet_by_id(self.ws_current_id)

    ## @param sheet_index Integer index of the sheet
    ## @param range_ Range object to get OdfCell. If None returns all OdfCell from sheet
    ## @return Returns a list of rows of object values
    def cells(self, sheet_index, range_=None):
        if range_ is None:
            range_=self.getSheetRange(sheet_index)
        else:
            range_=Range.assertRange(range_)
        r=[]
        for row in range(range_.numRows()):
            tmprow=[]
            for column in range(range_.numColumns()):
                tmprow.append(self.getCell(sheet_index, range_.start.addRowCopy(row).addColumnCopy(column)))
            r.append(tmprow)
        return r
        
    ## Returns a list of rows with the values of the sheet
    ## @param sheet_index Integer index of the sheet
    ## @param skip_up int. Number of rows to skip at the begining of the list of rows (lor)
    ## @param skip_down int. Number of rows to skip at the end of the list of rows (lor)
    ## @return Returns a list of rows of object values
    def values(self, sheet_index, skip_up=0, skip_down=0):
        r=[]
        for row in range(skip_up, self.rowNumber(sheet_index)-skip_down):
            tmprow=[]
            for column in range(self.columnNumber(sheet_index)):
                tmprow.append(self.getCellValue(sheet_index, Coord("A1").addRow(row).addColumn(column)))
            r.append(tmprow)
        return r
 
    ## @param sheet_index Integer index of the sheet
    ## @param range_ Range object to get values. If None returns all values from sheet
    ## @return Returns a list of rows of object values
    def values_by_range(self, sheet_index, range_):
        range_=Range.assertRange(range_)
        r=[]
        for row in range(range_.numRows()):
            tmprow=[]
            for column in range(range_.numColumns()):
                tmprow.append(self.getCellValue(sheet_index, range_.start.addRowCopy(row).addColumnCopy(column)))
            r.append(tmprow)
        return r
    
    ## @param sheet_index Integer index of the sheet
    ## @param column_letter Letter of the column to get values
    ## @param skip Integer Number of top rows to skip in the result
    ## @return List of values
    def getColumnValues(self, sheet_index, column_letter, skip_up=0, skip_down=0):
        r=[]
        for row in range(skip_up, self.rowNumber(sheet_index)-skip_down):
            r.append(self.getCellValue(sheet_index, Coord(column_letter+"1").addRow(row)))
        return r    

    ## @param sheet_index Integer index of the sheet
    ## @param row_number String Number of the row to get values
    ## @param skip Integer Number of top rows to skip in the result
    ## @return List of values
    def getRowValues(self, sheet_index, row_number, skip_left=0, skip_right=0):
        r=[]
        for column in range(skip_left, self.columnNumber(sheet_index)-skip_right):
            r.append(self.getCellValue(sheet_index, Coord("A"+row_number).addColumn(column)))
        return r

    ## Return a Range object with the limits of the index sheet
    def getSheetRange(self, sheet_index):
        self.setCurrentSheet(sheet_index)
        return Range(self.ws_current.calculate_dimension())
        
    def rowNumber(self, sheet_index):
        self.setCurrentSheet(sheet_index)
        return self.ws_current.max_row
        
    def columnNumber(self, sheet_index):
        self.setCurrentSheet(sheet_index)
        return self.ws_current.max_column
        
    ## Returns the cell value
    def getCellValue(self, sheet_index, coord):
        self.setCurrentSheet(sheet_index)
        cell=self.getCell(sheet_index, coord)
        return self.getCellValue_from_openpyxl_cell(cell)

    ## Returns the cell value
    def getCellValue_from_openpyxl_cell(self, cell):
        def hasCurrency(cell):
            for currency in ["EUR", "USD","GBP"]:
                if currency in cell.number_format and cell.data_type=="n":
                    return Currency(cell.value, currency)
            return None
        #----------------------
        #print(cell.number_format, cell.value, cell.data_type)
        #print(dir(cell))
        has_currency=hasCurrency(cell)
        if has_currency is not None:
            return has_currency
        elif cell.data_type=="f":#Formula
            return cell.value
        elif "%" in cell.number_format:
            return Percentage(cell.value, 1)
        elif cell.data_type=="d" and "H" not in cell.number_format.upper():#Date
            return cell.value.date()
        elif cell.data_type=="d" and "Y" not in cell.number_format.upper():#Tiime
            return cell.value
        elif cell.data_type=="d":#Datetime
            return cell.value
        elif cell.data_type == "b": #Boolean
            return cell.value
        elif cell.number_format=='"BOOL"e"AN"':
            if cell.value==1:
                return True
            elif cell.value==0:
                return False
        elif cell.data_type == "s":
            return str(cell.value)
        elif cell.value is None:
            return None
        elif cell.data_type=="n":
            try:
                return Decimal(str(cell.value))
            except:
                return str(cell.value)

    def getCell(self, sheet_index,  coord):
        self.setCurrentSheet(sheet_index)
        coord=Coord.assertCoord(coord)
        return self.ws_current[coord.string()]

class XLSX_Write(XLSX_Commons):
    ## @param filename
    ## @param template
    ## @param data_only load_workbook now accepts data_only to allow extracting values only from formulae. Default is false
    def __init__(self,filename, template=None, data_only=False, keep_vba=False):
        XLSX_Commons.__init__(self)
        self.filename=filename
        self.template=template
        if template is None:
            self.wb=openpyxl.Workbook()
        else:
            self.wb=openpyxl.load_workbook(self.template, keep_vba=keep_vba, data_only=data_only)

        self.ws_current=self.wb.active
        self.setCurrentSheet(self.ws_current.title)

        self.stOrange=openpyxl.styles.Color('FFFFDCA8')
        self.stYellow=openpyxl.styles.Color('FFFFFFC0')
        self.stGreen=openpyxl.styles.Color('FFC0FFC0')
        self.stGrayLight=openpyxl.styles.Color('FFDCDCDC')
        self.stGrayDark=openpyxl.styles.Color('FFC3C3C3')
        self.stWhite=openpyxl.styles.Color('FFFFFFFF')
    
    ## Returns the style name of a givenven color
    ## @param openpyxl.styles.Color
    ## @return string
    def styleName(self, color):
        if color==self.stOrange:
            return "Orange"
        elif color==self.stYellow:
            return "Yellow"
        elif color==self.stGreen:
            return "Green"
        elif color==self.stGrayDark:
            return "Dark gray"
        elif color==self.stGrayLight:
            return "Light gray"
        elif color==self.stWhite:
            return "White"
        elif color==None:
            return "Normal"

    ## Freeze panels in a sheet and sets the selected cell
    ## Selects a cell https://openpyxl.readthedocs.io/en/latest/_modules/openpyxl/worksheet/views.html#Selection
    ## To DEBUG THIS function rename xlsx to zip and enter in xl/worksheet to see code and props
    ## @param freeze_coord, Cell where panels are frrozen. Can be a string or a Coord object.
    ## @param selected_coord. Cell selected opening sheet. Can be a string or a Coord object.
    ## @param topLeftCell, topleftcell to show in sheet after opening. Can be a string or a Coord object.
    def freezeAndSelect(self, freeze_coord, selected_coord=None, topleftcell_coord=None):
        if selected_coord is None:
            selected_coord=Coord_from_index(self.columnNumber(self.ws_current_id)-1, self.rowNumber(self.ws_current_id)-1)

        if topleftcell_coord is None:
            topleftcell_coord=topLeftCellNone(freeze_coord, selected_coord)

        freeze_coord=Coord.assertCoord(freeze_coord)
        selected_coord=Coord.assertCoord(selected_coord)
        topleftcell_coord=Coord.assertCoord(topleftcell_coord)
        self.ws_current.freeze_panes=self.ws_current[freeze_coord.string()]
        sheet=self.ws_current.views.sheetView[0]
        if freeze_coord.letterIndex()>0 and freeze_coord.numberIndex()>0:#Freeze C3 WORKS
            sheet.selection[2].activeCell=selected_coord.string()
            sheet.selection[2].sqref=selected_coord.string()
            sheet.pane.activePane = 'bottomRight'
            sheet.pane.topLeftCell=topleftcell_coord.string()
        elif freeze_coord.letterIndex()==0 and freeze_coord.numberIndex()>0:#Freeze A3  WORKS
            topLeftCellFirst=topleftcell_coord.letter+"1"
            sheet.topLeftCell=topLeftCellFirst
            sheet.view="normal"
            sheet.pane.xSplit="0"
            sheet.selection[0].pane="bottomLeft"
            sheet.selection[0].activeCell=selected_coord.string()
            sheet.selection[0].sqref=selected_coord.string()
            sheet.selection[0].state="frozen"
            sel = list(sheet.selection)
            sel.insert(0, openpyxl.worksheet.worksheet.Selection(pane="topLeft", activeCell=topLeftCellFirst, sqref=topLeftCellFirst))
            sheet.selection = sel
            sheet.pane.topLeftCell=topleftcell_coord.string()
        elif freeze_coord.letterIndex()>0 and freeze_coord.numberIndex()==0:#Freeze C1 WORKS
            #Comparing output with a officegenerator good xlsx
            topLeftCellFirst="A"+topleftcell_coord.number
            sheet.topLeftCell=topLeftCellFirst
            sheet.pane.ySplit="0"
            sheet.selection[0].activeCell=selected_coord.string()
            sheet.selection[0].sqref=selected_coord.string()
            sel = list(sheet.selection)
            sel.insert(0, openpyxl.worksheet.worksheet.Selection(pane="topLeft", activeCell=topLeftCellFirst, sqref=topLeftCellFirst))
            sheet.selection = sel
            sheet.pane.topLeftCell=topleftcell_coord.string()
        elif freeze_coord.letterIndex()==0 and freeze_coord.numberIndex()==0:#Freeze A1 WORKS
            sheet.selection[0].activeCell=selected_coord.string()
            sheet.selection[0].sqref=selected_coord.string()
            sheet.selection[0].pane='topLeft'
            sheet.topLeftCell=topleftcell_coord.string()

    ## Changes name of the current sheet
    def setSheetName(self, name):
        self.ws_current.title=name

    ## Create a sheet at the end, renames it and selects it as current
    def createSheet(self, name):
        self.wb.create_sheet(title=name)
        self.setCurrentSheet(name)

    def setColorScale(self, range):
        self.ws_current.conditional_formatting.add(range, 
                            openpyxl.formatting.rule.ColorScaleRule(
                                                start_type='percentile', start_value=0, start_color='00FF00',
                                                mid_type='percentile', mid_value=50, mid_color='FFFFFF',
                                                end_type='percentile', end_value=100, end_color='FF0000'
                                                )
                                            )
    ## Returns sheet_name
    def sheet_name(self, id=None):
        if id==None:
            id=self.ws_current_id
        return self.wb.sheetnames[id]

    ## After removing it sets current sheet to 0 index
    def remove_sheet_by_id(self, id):
        ws=self.get_sheet_by_id(id)
        self.wb.remove(ws)
        self.setCurrentSheet(0)

    def save(self, filename=None):
        if filename==None:
            filename=self.filename
        if path.dirname(filename)!="":
            makedirs(path.dirname(filename), exist_ok=True)
        self.wb.save(filename)

        if path.exists(filename)==False:
            print(_("*** ERROR: File wasn't generated ***"))

    ## Internal function to set the number format
    ##
    ## This strings are openpyxl string not libreoffice cell string
    ## @param cell is a cell object
    ## @param value Value to add to the cell
    ## @param style Color or None. If None this function it's ignored
    ## @param decimals Number of decimals
    def __setNumberFormat(self, cell, value, style, decimals):     
        if style==None:
            return
        if value.__class__.__name__ in ("int", ):#Un solo valor
            cell.number_format='#,##0;[RED]-#,##0'
        elif value.__class__.__name__ in ("float", "Decimal"):#Un solo valor
            zeros=decimals*"0"
            cell.number_format="#,##0.{0};[RED]-#,##0.{0}".format(zeros)
        elif value.__class__.__name__ in ("datetime", ):
            cell.number_format="YYYY-MM-DD HH:mm"
        elif value.__class__.__name__ in ("date", ):
            cell.number_format="YYYY-MM-DD"
        elif value.__class__.__name__ in ("Currency", "Money" ):
            cell.number_format='#,##0.00 "{0}";[RED]-#,##0.00 "{0}"'.format(currency_symbol(value.currency))
        elif value.__class__.__name__ in ("Percentage",  ):
            cell.number_format="#.##0,00 %;[RED]-#.##0,00 %"
        elif value.__class__.__name__ =="bool":
            cell.number_format="BOOLEAN"

    ## Internal function to set the number format of a formula
    ##
    ## This strings are openpyxl string not libreoffice cell string
    ## @param cell is a cell object
    ## @param resultclass int, float, Decimal, datetime.datetime, datetime.date,"€","$" (Currency.symbol), Percentage
    ##    Currency it's not used because value is a string, and I can't get value.currency. Maybe I should create a Formula object, but Iwill have to change a lot of apps
    ## @param style Color or None. If None this function it's ignored
    ## @param decimals Number of decimals
    def __setFormulaNumberFormat(self, cell, value, resultclass, style, decimals):     
        if style==None:
            return
        if resultclass in ("int", ):#Un solo valor
            cell.number_format='#,##0;[RED]-#,##0'
        elif resultclass in ("float", "Decimal"):#Un solo valor
            zeros=decimals*"0"
            cell.number_format="#,##0.{0};[RED]-#,##0.{0}".format(zeros)
        elif resultclass in ("datetime", ):
            cell.number_format="YYYY-MM-DD HH:mm"
        elif resultclass in ("date", ):
            cell.number_format="YYYY-MM-DD"
        elif resultclass.__class__ in ("str", ):
            cell.number_format='#,##0.00 "{0}";[RED]-#,##0.00 "{0}"'.format(resultclass)
        elif resultclass in ("Percentage", ):
            cell.number_format="#.##0,00 %;[RED]-#.##0,00 %"
        elif resultclass in ("$", "€"):
            cell.number_format='#,##0.00 "{0}";[RED]-#,##0.00 "{0}"'.format(resultclass)
        elif resultclass=="bool":
            cell.number_format="BOOLEAN"

    ## Internat function to set a cell. All properties except border that it's setted in overwrite functions (merged and no merged)
    ## @param cell is a cell object
    def __setValue(self, cell, value, style, decimals, alignment):     
        if value==None:
            return
        elif value.__class__.__name__ in ("Currency", "Money" ):
            cell.value=value.amount
        elif value.__class__.__name__ in ("Percentage", ):
            cell.value=value.value
        elif value.__class__.__name__ == "bool":
            cell.value=value
        else:
            cell.value=value

    ## Internal method to set a not merged cell
    ## @param cell is a cell object
    ## @param value Value to add to the cell
    ## @param style Color or None. If None this function it's ignored
    ## @param decimals Number of decimals
    ## @param alignment Cell alignment
    def __setCell(self, coord, value, style=None, decimals=2, alignment=None):
        coord=Coord.assertCoord(coord)
        cell=self.getCell(self.ws_current_id, coord.string())
        self.__setValue(cell, value, style, decimals, alignment)
        self.__setBorder(cell, style)
        self.__setAlignment(cell, value, style, alignment)
        self.__setNumberFormat(cell, value, style, decimals)      

        if style!=None:
            cell.fill=openpyxl.styles.PatternFill("solid", fgColor=style)
            bold=False if style==self.stWhite else True
            cell.font=openpyxl.styles.Font(name='Arial', size=10, bold=bold)

    ## Internat function to set cell alignment
    ## @param cell is a cell object
    ## @style Color or None. This method is ignored if style=None
    def __setAlignment(self, cell, value,  style, alignment):  
        if style==None:
            return
        if alignment==None:
            if value.__class__.__name__ in ("str", ):#Un solo valor
                if value.startswith("=") or value.startswith("+"):
                    alignment='right'
                else:
                    alignment='left'
            else:
                alignment='right'
        cell.alignment=openpyxl.styles.Alignment(horizontal=alignment, vertical='center')

    ## Writes a cell or a list of cell or a list of list of cells
    ## @param coord Can be a Coord or a string with text coords
    ## @param result Can be a value, a list of values or a list of lists of values
    ## @param style its a openpyxl.styles.Color object. There are several predefined stGreen, stGrayDark, stGrayLight, stOrange, stYellow, stWhite or None. None is used to preserve template cell and the value is the only thing will be changed
    ## @param decimals Integer with the number of decimals. 2 by default
    ## @param alignment String None by default. Can be "right","left","center"
    def overwrite(self, coord, result, style=None,  decimals=2, alignment=None):
        coord=Coord.assertCoord(coord)
        if result.__class__== list:#Una lista
            for i,row in enumerate(result):
                if row.__class__ in (list, ):#Una lista de varias columnas
                    for j,column in enumerate(row):
                        self.__setCell(Coord(coord.string()).addRow(i).addColumn(j), result[i][j], style, decimals, alignment )   
                else:#Una lista de una columna
                    self.__setCell(Coord(coord.string()).addRow(i), result[i], style, decimals, alignment )
        else:#Un solo valor
            self.__setCell(coord, result, style, decimals, alignment )
            

    ## Writes a formula in a cell 
    ## @param coord Can be a Coord or a string with text coord
    ## @param result Can be a value
    ## @param resultclass string that can be: int, float, Decimal, datetime.datetime, datetime.date, Currency, Percentage
    ## @param style its a openpyxl.styles.Color object. There are several predefined stGreen, stGrayDark, stGrayLight, stOrange, stYellow, stWhite or None. None is used to preserve template cell and the value is the only thing will be changed
    ## @param decimals Integer with the number of decimals. 2 by default
    ## @param alignment String None by default. Can be "right","left","center"
    def overwrite_formula(self, coord, value, resultclass=None, style=None, decimals=2, alignment=None):
        if isFormula(value)==False:
            print(_("This is not a formula. You can't use overwrite_formula"))
            return
        if value.__class__==list:
            print("Adding formula list is not allowed")
            return
        coord=Coord.assertCoord(coord)
        cell=self.getCell(self.ws_current_id, coord)       
        
        self.__setValue(cell, value, style, decimals, alignment)
        self.__setBorder(cell, style)
        self.__setAlignment(cell, value, style, alignment)
        self.__setFormulaNumberFormat(cell, value, resultclass, style, decimals)      

        if style!=None:
            cell.fill=openpyxl.styles.PatternFill("solid", fgColor=style)
            bold=False if style==self.stWhite else True
            cell.font=openpyxl.styles.Font(name='Arial', size=10, bold=bold)
            
            
    ## @param cood Coord from we are going to add totals
    ## @param list_of_totals List with strings or keys. Example: ["Total", "#SUM", "#AVG"]...
    ## @param list_of_styles List with string styles or None. If none tries to guest from top column object. List example: ["GrayLightPercentage", "GrayLightInteger"]
    ## @param string with the row where th3e total begins
    ## @param string with the rew where the formula ends. If None it's a coord.row -1
    def overwriteTotalsHorizontal(self, coord, list_of_totals, list_of_styles=None, row_from="2", row_to=None,  list_of_decimals=2):
        coord=Coord.assertCoord(coord)
        for letter, total in enumerate(list_of_totals):
            coord_total=coord.addColumnCopy(letter)
            coord_total_from=Coord(coord_total.letter+row_from)
            if row_to is None:
                coord_total_to=coord_total.addRowCopy(-1)# row above
            else:
                coord_total=Coord(coord_total.letter+row_to)

            if list_of_styles is None:
                style=self.stGrayLight
            else:
                style=list_of_styles[letter]
                
            if list_of_decimals.__class__.__name__=="int":
                decimals=2
            else:
                decimals=list_of_decimals[letter]

            self.overwrite(coord_total, generate_formula_total_string(total, coord_total_from, coord_total_to),  style, decimals)
            
    ## @param cood Coord from we are going to add totals
    ## @param list_of_totals List with strings or keys. Example: ["Total", "#SUM", "#AVG"]...
    ## @param list_of_styles List with string styles or None. If none tries to guest from top column object. List example: ["GrayLightPercentage", "GrayLightInteger"]
    ## @param string with the row where th3e total begins
    ## @param string with the rew where the formula ends. If None it's a coord.row -1
    def overwriteTotalsVertical(self, coord, list_of_totals, list_of_styles=None, column_from="B", column_to=None,  list_of_decimals=2):
        coord=Coord.assertCoord(coord)
        for i, total in enumerate(list_of_totals):
            coord_total=coord.addRowCopy(i)
            coord_total_from=Coord(column_from+coord_total.number)
            if column_to is None:
                coord_total_to=coord_total.addColumnCopy(-1)# row above
            else:
                coord_total=Coord(column_to+coord_total.number)

            if list_of_styles is None:
                style=self.stGrayLight
            else:
                style=list_of_styles[i]

            if list_of_decimals.__class__.__name__=="int":
                decimals=2
            else:
                decimals=list_of_decimals[i]


            self.overwrite(coord_total, generate_formula_total_string(total, coord_total_from, coord_total_to),  style, decimals)


    ##Sets border to a cell not merged
    ## @param cell is a cell object
    ## @param style Color or None. If None this function it's ignored
    def __setBorder(self, cell, style):
        if style==None:
            return
        cell.border=openpyxl.styles.Border(
            left=openpyxl.styles.Side(border_style='thin'),
            top=openpyxl.styles.Side(border_style='thin'),
            right=openpyxl.styles.Side(border_style='thin'),
            bottom=openpyxl.styles.Side(border_style='thin') 
        )

    ## Sets cell name to use in formulas. Fails if range_string is not fixed. For example: $A$4
    ## @param range_string 
    ## @param name
    def setCellName(self, range_string, name):
        self.wb.create_named_range(name, self.ws_current, range_string)

    ## Set columns width in current sheet
    ## @param arrWidths List with integers representing column width
    def setColumnsWidth(self, arrWidths):
        for i in range(len(arrWidths)):
            self.ws_current.column_dimensions[columnAdd("A", i)].width=arrWidths[i]

    ## Create a merged cell
    ## @param range_string Can be a Range or a range string 
    ## @param value Can be a value. Must be the value only for the first cell
    ## @param style its a openpyxl.styles.Color object. There are several predefined stGreen, stGrayDark, stGrayLight, stOrange, stYellow, stWhite or None. None is used to preserve template cell and the value is the only thing will be changed
    ## @param decimals Integer with the number of decimals. 2 by default
    ## @param alignment String None by default. Can be "right","left","center"
    def overwrite_and_merge(self, range,  value, style=None,  decimals=2, alignment=None):
        range=Range.assertRange(range)
        self.ws_current.merge_cells(range.string())
        top = openpyxl.styles.Border(top=openpyxl.styles.Side(border_style='thin'))
        left = openpyxl.styles.Border(left=openpyxl.styles.Side(border_style='thin'))
        right = openpyxl.styles.Border(right=openpyxl.styles.Side(border_style='thin'))
        bottom = openpyxl.styles.Border(bottom=openpyxl.styles.Side(border_style='thin'))

        self.__setCell(range.start.string(), value, style, decimals, alignment)

        rows = self.ws_current[range.string()]

        for cell in rows[0]:
            cell.border = cell.border + top
        for cell in rows[-1]:
            cell.border = cell.border + bottom

        for row in rows:
            l = row[0]
            r = row[-1]
            l.border = l.border + left
            r.border = r.border + right

    ## Sets a comment
    ## @param strcell String "A1" for example
    def setComment(self, coord_string, comment):
        self.ws_current[coord_string].comment=openpyxl.comments.Comment(comment, "PySGAE")

class XLSX_Read(XLSX_Commons):
    ## @param filename
    ## @param data_only load_workbook now accepts data_only to allow extracting values only from formulae. Default is False to avoid lose information
    ## @param keep_vba. Default is True
    def __init__(self, filename, data_only=False, keep_vba=True):
        XLSX_Commons.__init__(self)
        self.filename=filename
        self.wb=openpyxl.load_workbook(self.filename, keep_vba=keep_vba, data_only=data_only)
        self.ws_current=self.wb.active
        self.setCurrentSheet(self.ws_current.title)


## Gets a XLSX file and rewrites it with libreoffice convert-to command
## Can be used to assign data values formulas to file. Or to fix ploblems on specific files.
## I had problems with concurrency due to convrert-to opens a file and locks it. I should not use convert-to with concurrency
## @param filename_from canbe an ods or a xlsx file.
## @param filename_to Must be a xlsx file path
## Returns the name of the recently created file
def create_rewritten_xlsx(filename_from, filename_to=None):
    if filename_to is None:
        filename_to=filename_from + ".rewritten.xlsx"
    convert_command(filename_from, filename_to)
    return filename_to


## @param filename_from canbe an ods or a xlsx file.
## @param filename_to Must be a xlsx file path
## Returns the name of the recently created file
def create_data_only_xlsx(filename_from, filename_to=None):
    if filename_to is None:
        filename_to=filename_from + ".data_only.xlsx"
    with TemporaryDirectory(prefix="officegenerator_") as tmp_name:
        temporal_path="{}/{}.rewritten.xlsx".format(tmp_name, path.basename(filename_to))
        convert_command(filename_from, temporal_path)
        xlsx=XLSX_Write(filename_to, template=temporal_path, data_only=True)
        xlsx.save()
    return filename_to

## Creates a new file, from another file. It automatically rewrites it to obtein values, and sorts a sheet.
## @param filename_from canbe an ods or a xlsx file.
## @param filename_to Must be a xlsx file path
## @param sheet_index
## @param column_index
## @param reverse
## @param skip_up, to skip headers by default 1
## @param skip_down to skip headers by default 0
## Returns the name of the recently created file
def create_sorting_sheet(filename_from, filename_to, sheet_index, column_index, reverse=False,  skip_up=1, skip_down=0):
    with TemporaryDirectory(prefix="officegenerator_") as tmp_name:
        create_rewritten_xlsx(filename_from , tmp_name + sep + "rewritten.xlsx")
        xlsx_read=XLSX_Read(tmp_name + sep + "rewritten.xlsx")
        rows=xlsx_read.cells(sheet_index)
        value_rows=rows[skip_up:len(rows)-skip_down]
        value_rows=sorted(value_rows, key=lambda c: xlsx_read.getCellValue_from_openpyxl_cell(c[column_index]), reverse=reverse)
        
        xlsx_write=XLSX_Write(tmp_name + sep + "sorted.xlsx", template=tmp_name + sep + "rewritten.xlsx")
        xlsx_write.setCurrentSheet(sheet_index)
        for number_index, row in enumerate(value_rows):
            for letter_index, cell in enumerate(row):
                xlsx_write.ws_current.cell(row=1+number_index+skip_up, column=1+ letter_index).value=xlsx_read.getCellValue_from_openpyxl_cell(cell)         
        xlsx_write.save()        
        
        convert_command(tmp_name + sep + "sorted.xlsx", filename_to)
        



## @namespace officegenerator.tests.test_commons
## @brief Test for officegenerator.commons functions and classes

import unittest
from officegenerator.commons import Range, Coord


## Class to text Range operations. Class must begin with Test and modules with test_ too

class TestCoord(unittest.TestCase):
    def test_Coord_methods(self):
        s="Z1"
        self.assertEqual(Coord(s).addColumn().string(), "AA1")
        self.assertEqual(Coord(s).addColumn(-25).string(), "A1")
        self.assertEqual(Coord(s).addRow(25).string(), "Z26")
        self.assertEqual(Coord(s).letterIndex(), 25)
        self.assertEqual(Coord(s).letterPosition(), 26)
        self.assertEqual(Coord(s).numberIndex(), 0)
        self.assertEqual(Coord(s).numberPosition(), 1)
        self.assertEqual(Coord(s).string(),"Z1")
        self.assertEqual(str(Coord(s)),"Coord <Z1>")

    def test_Coord_methods_in_the_limit(self):
        s="Z1"
        self.assertEqual(Coord(s).addColumn(-26).string(), "A1")
        self.assertEqual(Coord(s).addColumn(-2600).string(), "A1")
        self.assertEqual(Coord(s).addRow(-1).string(), "Z1")

class TestRange(unittest.TestCase):
    def test_Range_methods(self):
        s="A1:B2"
        r="Z10:AA20"
        #Normal use, making range bigger
        self.assertEqual(Range(s).addColumnAfter(2).string(),"A1:D2")
        self.assertEqual(Range(s).addRowAfter(2).string(),"A1:B4")
        self.assertEqual(Range(r).addColumnBefore(2).string(),"X10:AA20")
        self.assertEqual(Range(r).addRowBefore(2).string(),"Z8:AA20")

        self.assertEqual(Range(s).numColumns(), 2)
        self.assertEqual(Range(s).numRows(), 2)
        self.assertEqual(Range(s).string(), s)

    def test_Range_methods_in_the_limit(self):
        s="A1:B2"
        r="Z10:AA20"
        #Negative use, cutting range
        self.assertEqual(Range(s).addColumnAfter(-2).string(),"A1:A2")
        self.assertEqual(Range(s).addRowAfter(-2).string(),"A1:B1")
        self.assertEqual(Range(r).addColumnBefore(-2).string(),"AB10:AA20")
        self.assertEqual(Range(r).addRowBefore(-2).string(),"Z12:AA20")

if __name__ == '__main__':
    unittest.main()


from setuptools import setup, Command
import site
import os
import platform



class Reusing(Command):
    description = "Download modules from https://github.com/turulomio/reusingcode/"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from sys import path
        path.append("officegenerator")
        from github import download_from_github
        download_from_github('turulomio','reusingcode','python/github.py', 'officegenerator')
        download_from_github('turulomio','reusingcode','python/casts.py', 'officegenerator')
        download_from_github('turulomio','reusingcode','python/datetime_functions.py', 'officegenerator')
        download_from_github('turulomio','reusingcode','python/decorators.py', 'officegenerator')
        download_from_github('turulomio','reusingcode','python/libmanagers.py', 'officegenerator')
        download_from_github('turulomio','reusingcode','python/objects/percentage.py', 'officegenerator/objects/')
        download_from_github('turulomio','reusingcode','python/objects/currency.py', 'officegenerator/objects/')

## Class to define doc command
class Doc(Command):
    description = "Update translations"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        #es
        os.system("xgettext -L Python --no-wrap --no-location --from-code='UTF-8' -o locale/officegenerator.pot *.py officegenerator/*.py")
        os.system("msgmerge -N --no-wrap -U locale/es.po locale/officegenerator.pot")
        os.system("msgfmt -cv -o officegenerator/locale/es/LC_MESSAGES/officegenerator.mo locale/es.po")

class Procedure(Command):
    description = "Show release procedure"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print("""Nueva versión:
  * Cambiar la versión y la fecha en commons.py
  * Modificar el Changelog en README
  * python setup.py doc
  * linguist
  * python setup.py doc
  * python setup.py install
  * python setup.py doxygen
  * git commit -a -m 'officegenerator-{}'
  * git push
  * Hacer un nuevo tag en GitHub
  * python setup.py sdist upload -r pypi
  * python setup.py uninstall
  * Crea un nuevo ebuild de Gentoo con la nueva versión
  * Subelo al repositorio del portage
""".format(__version__))

## Class to define doxygen command
class Doxygen(Command):
    description = "Create/update doxygen documentation in doc/html"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print("Creating Doxygen Documentation")
        os.system("""sed -i -e "41d" doc/Doxyfile""")#Delete line 41
        os.system("""sed -i -e "41iPROJECT_NUMBER         = {}" doc/Doxyfile""".format(__version__))#Insert line 41
        os.system("rm -Rf build")
        os.chdir("doc")
        os.system("doxygen Doxyfile")
        os.system("rsync -avzP -e 'ssh -l turulomio' html/ frs.sourceforge.net:/home/users/t/tu/turulomio/userweb/htdocs/doxygen/officegenerator/ --delete-after")
        os.chdir("..")

## Class to define uninstall command
class Uninstall(Command):
    description = "Uninstall installed files with install"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        if platform.system()=="Linux":
            os.system("rm -Rf {}/officegenerator*".format(site.getsitepackages()[0]))
            os.system("rm /usr/bin/officegenerator*")
        else:
            os.system("pip uninstall officegenerator")

########################################################################

## Version of officegenerator captured from commons to avoid problems with package dependencies
__version__= None
with open('officegenerator/commons.py', encoding='utf-8') as f:
    for line in f.readlines():
        if line.find("__version__ =")!=-1:
            __version__=line.split("'")[1]


setup(name='officegenerator',
     version=__version__,
     description='Python module to read and write LibreOffice and MS Office files',
     long_description='Project web page is in https://github.com/turulomio/officegenerator',
     long_description_content_type='text/markdown',
     classifiers=['Development Status :: 4 - Beta',
                  'Intended Audience :: Developers',
                  'Topic :: Software Development :: Build Tools',
                  'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
                  'Programming Language :: Python :: 3',
                 ], 
     keywords='office generator',
     url='https://officegenerator.sourceforge.io/',
     author='Turulomio',
     author_email='turulomio@yahoo.es',
     license='GPL-3',
     packages=['officegenerator'],
     install_requires=['odfpy==1.3.6','openpyxl'],
     entry_points = {'console_scripts': ['officegenerator_demo=officegenerator.demo:main',
                                         'officegenerator_odf2xml=officegenerator.odf2xml:main',
                                         'officegenerator_xlsx2xml=officegenerator.xlsx2xml:main',
                                        ],
                    },
     cmdclass={'doxygen': Doxygen,
               'uninstall':Uninstall, 
               'doc': Doc,
               'procedure': Procedure,
               'reusing': Reusing,
              },
     zip_safe=False,
     test_suite = 'officegenerator.tests',
     include_package_data=True
)



## @namespace officegenerator.tests.test_libodfgenerator
## @brief Test for officegenerator.libodfgenerator functions and classes
import datetime
import unittest
from decimal import Decimal
from officegenerator.commons import Percentage, Currency
from officegenerator.demo import demo_ods
from officegenerator.libodfgenerator import ODS_Read, guess_ods_style

## Class to text Range operations. Class must begin with Test and modules with test_ too
class TestFunctions(unittest.TestCase):
    def test_guess_ods_style(self):
        s=Percentage(1,2)
        self.assertEqual(guess_ods_style("Orange", s), "OrangePercentage")

## Class to text ODS_Read operations. Class must begin with Test and modules with test_ too
class TestODS_Read(unittest.TestCase):
    def test_guess_ods_style(self):
        demo_ods()
        doc=ODS_Read("officegenerator.ods")
        s1=doc.getSheetElementByIndex(0)
        self.assertEqual(doc.getCellValue(s1,"A2"),"Percentage")
        #self.assertEqual(doc.getCellValue(s1,"B2").value, Percentage(21.43,100).value)
        self.assertEqual(doc.getCellValue(s1,"B4").upper(),"=SUM(B2:B3)")
        self.assertEqual(doc.getCellValue(s1,"B6"), Decimal("100.26"))

        s3=doc.getSheetElementByIndex(2)
        self.assertEqual(doc.getCellValue(s3,"B2").__class__, datetime.datetime)
        self.assertEqual(doc.getCellValue(s3,"C2").__class__, datetime.date)
        self.assertEqual(doc.getCellValue(s3,"E2").string(), Currency(12.56, "EUR").string())

if __name__ == '__main__':
    unittest.main()



## @namespace officegenerator.tests.test_libxlsxgenerator
## @brief Test for officegenerator.libxlsxgenerator functions and classes
import unittest
from officegenerator.libxlsxgenerator import XLSX_Write

## Class to text XLSX_Write methods. Class must begin with Test and modules with test_ too
class TestXLSX_Write(unittest.TestCase):
    ## It creates and removes sheets
    def test_create_remove(self):
         xlsx=XLSX_Write("officegenerator.xlsx")
         xlsx.setCurrentSheet(0)
         xlsx.setSheetName("It was")
         xlsx.createSheet("New")
         xlsx.remove_sheet_by_id(0)
         self.assertEqual(xlsx.sheet_name(0),"New")

if __name__ == '__main__':
    unittest.main()



## @namespace officegenerator.demo
## @brief Generate ODF example files

import argparse
import gettext
import pkg_resources
from concurrent.futures import ProcessPoolExecutor, as_completed
from datetime import timedelta, datetime, date
from decimal import Decimal
from multiprocessing import cpu_count

from officegenerator.commons import __version__, addDebugSystem
from officegenerator.libodfgenerator import ODS_Read, ODS_Write, ODT_Manual_Styles, ODT_Standard,  OdfCell, ColumnWidthODS, ODT, create_rewritten_ods, create_data_only_ods
from officegenerator.libxlsxgenerator import XLSX_Write, XLSX_Read, create_data_only_xlsx, create_rewritten_xlsx, create_sorting_sheet
from officegenerator.standard_sheets import Model, Model_Auto
from officegenerator.commons import argparse_epilog, Coord, Range
from officegenerator.objects.currency import Currency
from officegenerator.objects.percentage import Percentage
from odf.text import P
import openpyxl.styles
from os import remove

try:
    t=gettext.translation('officegenerator',pkg_resources.resource_filename("officegenerator","locale"))
    _=t.gettext
except:
    _=str

def remove_without_errors(filename):
    try:
        remove(filename)
    except OSError as e:
        print(_("Error deleting: {} -> {}".format(filename, e.strerror)))

## If arguments is None, launches with sys.argc parameters. Entry point is toomanyfiles:main
## You can call with main(['--pretend']). It's equivalento to os.system('program --pretend')
## @param arguments is an array with parser arguments. For example: ['--argument','9']. 
def main(arguments=None):
    parser=argparse.ArgumentParser(prog='officegenerator', description=_('Create example files using officegenerator module'), epilog=argparse_epilog(), formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--debug', help="Debug program information", choices=["DEBUG","INFO","WARNING","ERROR","CRITICAL"], default="ERROR")
    group= parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--create', help="Create demo files", action="store_true",default=False)
    group.add_argument('--remove', help="Remove demo files", action="store_true", default=False)
    args=parser.parse_args(arguments)

    addDebugSystem(args.debug)

    if args.remove==True:
        remove_without_errors("officegenerator.ods")
        remove_without_errors("officegenerator.ods.rewritten.ods")
        remove_without_errors("officegenerator.odt")
        remove_without_errors("officegenerator_manual_styles.odt")
        remove_without_errors("officegenerator_search_and_replace.odt")
        remove_without_errors("officegenerator_updated.ods")
        remove_without_errors("officegenerator.xlsx")
        remove_without_errors("officegenerator.xlsx.rewritten.xlsx")
        remove_without_errors("officegenerator.xlsx.data_only.xlsx")
        remove_without_errors("officegenerator_xlsx.data_only.txt")
        remove_without_errors("officegenerator_updated.xlsx")
        remove_without_errors("officegenerator_xlsx_readonly.txt")
        remove_without_errors("officegenerator_ods_readonly.txt")
        remove_without_errors("officegenerator_standard_sheets.ods")
        remove_without_errors("officegenerator_standard_sheets.odt")
        remove_without_errors("officegenerator_standard_sheets.xlsx")
        remove_without_errors("officegenerator_standard_sheets_direct.ods")
        remove_without_errors("officegenerator_standard_sheets_direct.xlsx")

    if args.create==True:
        start=datetime.now()
        futures=[]
        with ProcessPoolExecutor(max_workers=cpu_count()+1) as executor:
            futures.append(executor.submit(demo_ods))
            futures.append(executor.submit(demo_odt_standard))
            futures.append(executor.submit(demo_odt_manual_styles))
            futures.append(executor.submit(demo_xlsx))
            futures.append(executor.submit(demo_standard_sheets))

        for future in as_completed(futures):
            print(future.result())
            
            
        create_rewritten_xlsx("officegenerator.xlsx")
        create_data_only_xlsx("officegenerator.xlsx")
        create_rewritten_ods("officegenerator.ods")
        create_data_only_ods("officegenerator.ods")
        create_sorting_sheet("officegenerator.xlsx", "officegenerator.sorted.xlsx", 5, 11)
        #create_data_only_xlsx("officegenerator.ods", "officegenerator_from_ods.data_only.xlsx")
            
            
        ## This process depend of above files
        futures=[]
        with ProcessPoolExecutor(max_workers=cpu_count()+1) as executor:
            futures.append(executor.submit(demo_ods_updated))
            futures.append(executor.submit(demo_ods_readonly))
            futures.append(executor.submit(demo_odt_search_and_replace))
            futures.append(executor.submit(demo_xlsx_updated))
            futures.append(executor.submit(demo_xlsx_readonly))
            futures.append(executor.submit(demo_xlsx_data_only))

        for future in as_completed(futures):
            print(future.result())
        print("All process took {}".format(datetime.now()-start))

def demo_ods_readonly():
    start=datetime.now()
    doc=ODS_Read("officegenerator.ods")
    output=open("officegenerator_ods_readonly.txt", "w")
    
    output.write("{} {}\n".format("A1",  doc.getCellValue(9, "A1")))        
    output.write("{}\n".format(doc.getColumnValues(1, "J", skip_up=150)))
    output.write("{}\n".format(doc.getRowValues(1, "100", skip_left=3)))
        
    output.write("{}\n".format(doc.values_by_range(9, "A2:K2"))   )
    output.close()
    return "demo_ods_readonly took {}".format(datetime.now()-start)

def demo_ods_updated():
    start=datetime.now()
    doc=ODS_Read("officegenerator.ods")

    #Sustituye celda
    odfcell=doc.getCell(0, "B6")
    odfcell.object=1789.12
    odfcell.setComment(_("This cell has been updated and modified"))
    doc.setCell(0, "B6", odfcell)

    #Added cell
    odfcell=doc.getCell(0, "B10")
    odfcell.object=_("Created cell")
    odfcell.setComment(_("This cell has been updated and modified"))
    doc.setCell(0, "B10", odfcell )

    doc.save("officegenerator_updated.ods")
    return "demo_ods_updated took {}".format(datetime.now()-start)

def demo_ods():
    start=datetime.now()
    doc=ODS_Write("officegenerator.ods")
    doc.setMetadata(_("OfficeGenerator ODS example"),  _("Demo with ODS_Write class"), "Turulomio", _("This file have been generated with OfficeGenerator-{}. You can see OfficeGenerator main page in http://github.com/turulomio/officegenerator").format(__version__), "officegenerator demo files")
    s1=doc.createSheet("Example")
    s1.add("A1", [["Title", "Value"]], "OrangeCenter")
    s1.add("A2", "Percentage", "YellowLeft")
    s1.add("A4",  "Suma", "WhiteCenter")
    s1.add("B2",  Percentage(12, 56), "WhitePercentage")
    s1.add("B3",  Percentage(12, 21), "WhitePercentage")
    s1.add("B4",  "=sum(B2:B3)","WhitePercentage" )
    s1.add("B6",  100.26, "WhiteDecimal6")
    s1.add("B7",  101, "WhiteInteger")
    s1.freezeAndSelect("A2", "A30", "A5")

    #Manual cell
    cell=OdfCell("B10", "Celda con OdfCell", "YellowCenter")
    cell.setComment("Comentario")
    cell.setSpanning(2, 2)
    s1.addCell(cell)
    
    #Better way
    s1.addMerged("E10:F11", "Celda con Merged", "GrayDarkCenter")
    s1.setComment("E10", _("This is a comment"))

    #Using lists in arr
    s1.add("A13", [["Una fila"]*3], "Orange")
    s1.add("A15", [[12.3]*3, [12.3]*3])


    sf1=doc.createSheet("Freeze A1")
    for letter in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
        for number in range(1, 200):
            sf1.add(letter + str(number), letter+str(number), "YellowLeft")
    sf1.freezeAndSelect("A1", "Z199","U180") 

    sf2=doc.createSheet("Freeze A3")
    for letter in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
        for number in range(1, 200):
            sf2.add(letter + str(number), letter+str(number), "YellowLeft")
    sf2.freezeAndSelect("A3", "Z199","M171") 

    sf3=doc.createSheet("Freeze C1")
    for letter in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
        for number in range(1, 200):
            sf3.add(letter + str(number), letter+str(number), "YellowLeft")
    sf3.freezeAndSelect("C1", "Z199","Q168") 

    sf4=doc.createSheet("Freeze C3")
    for letter in "ABCDEFG":
        for number in range(1, 200):
            sf4.add(letter + str(number), letter+str(number), "YellowLeft")
    sf4.freezeAndSelect("C3", "G199","Q168") 


    sf1=doc.createSheet("Freeze A1 None")
    for letter in "ABCDEFG":
        for number in range(1, 200):
            sf1.add(letter + str(number), letter+str(number), "YellowLeft")
    sf1.freezeAndSelect("A1", "G199") 

    sf2=doc.createSheet("Freeze A3 None")
    for letter in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
        for number in range(1, 15):
            sf2.add(letter + str(number), letter+str(number), "YellowLeft")
    sf2.freezeAndSelect("A3", "Z14") 

    sf3=doc.createSheet("Freeze C1 None")
    for letter in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
        for number in range(1, 15):
            sf3.add(letter + str(number), letter+str(number), "YellowLeft")
    sf3.freezeAndSelect("C1", "Z14") 

    sf4=doc.createSheet("Freeze C3 None")
    for letter in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
        for number in range(1, 200):
            sf4.add(letter + str(number), letter+str(number), "YellowLeft")
    sf4.freezeAndSelect("C3", "Z199") 
    s6=doc.createSheet("Format number")
    s6.setColumnsWidth([ColumnWidthODS.L, ColumnWidthODS.Datetime, ColumnWidthODS.Date, ColumnWidthODS.L, ColumnWidthODS.L, ColumnWidthODS.L, ColumnWidthODS.L, ColumnWidthODS.XL, ColumnWidthODS.L, ColumnWidthODS.L, ColumnWidthODS.L])

    s6.add("A1", _("Style name"), "OrangeCenter")
    s6.add("B1", _("Date and time"), "OrangeCenter")
    s6.add("C1", _("Date"), "OrangeCenter")
    s6.add("D1", _("Integer"), "OrangeCenter")
    s6.add("E1", _("Euros"), "OrangeCenter")
    s6.add("F1", _("Dollars"), "OrangeCenter")
    s6.add("G1", _("Percentage"), "OrangeCenter")
    s6.add("H1", _("Number with 2 decimals"), "OrangeCenter")
    s6.add("I1", _("Number with 6 decimals"), "OrangeCenter")
    s6.add("J1", _("Time"), "OrangeCenter")
    s6.add("K1", _("Boolean"), "OrangeCenter")
    for row, color in enumerate(doc.colors.arr):
        s6.add(Coord("A2").addRow(row), color.name, color.name + "Left")
        s6.add(Coord("B2").addRow(row), datetime.now(),  color.name +"Datetime")
        s6.add(Coord("C2").addRow(row), date.today(), color.name + "Date")
        s6.add(Coord("D2").addRow(row), pow(-1, row)*-10000000, color.name+ "Integer")
        s6.add(Coord("E2").addRow(row), Currency(pow(-1, row)*12.56, "EUR"), color.name + "EUR")
        s6.add(Coord("F2").addRow(row), Currency(pow(-1, row)*12345.56, "USD"), color.name + "USD")
        s6.add(Coord("G2").addRow(row), Percentage(pow(-1, row)*1, 3), color.name+"Percentage")
        s6.add(Coord("H2").addRow(row), pow(-1, row)*123456789.121212, color.name+"Decimal2")
        s6.add(Coord("I2").addRow(row), pow(-1, row)*-12.121212, color.name+"Decimal6")
        s6.add(Coord("J2").addRow(row), (datetime.now()+timedelta(seconds=3600*12*row)).time(), color.name+"Time")
        s6.add(Coord("K2").addRow(row), bool(row%2),  color.name+"Left")

    s6.setComment("B2", _("This is a comment"))
    
    #Merge cells
    s6.addMerged("B13:F14", _("This cell is going to be merged with B13 a F14"), "GreenCenter")
    s6.addMerged("B18:G18", _("This cell is going to be merged and aligned desde B18 a G18"), "YellowRight")
    s6.freezeAndSelect("A11", "B11", "A11")#Default values
    
    #Cells with formula with diferent styles and number formats
    s6.addMerged("A20:D20",  _("These cells show formulas with diferent styles and number formats"), "Green")
    s6.add("E20",  "=2+3.3", "NormalDecimal6")
    s6.add("F20",  "=2+3.3", "YellowEUR")
    s6.add("G20",  "=2+3.3", "GreenDatetime")
    s6.add("H20",  "=0.9/23", "WhitePercentage")


    s7=doc.createSheet("Add totals example")
    s7.add("A1", _("Style name"), "OrangeCenter")
    s7.add("B1", _("Date and time"), "OrangeCenter")
    s7.add("C1", _("Date"), "OrangeCenter")
    s7.add("D1", _("Integer"), "OrangeCenter")
    s7.add("E1", _("Euros"), "OrangeCenter")
    s7.add("F1", _("Dollars"), "OrangeCenter")
    s7.add("G1", _("Percentage"), "OrangeCenter")
    s7.add("H1", _("Number with 2 decimals"), "OrangeCenter")
    s7.add("I1", _("Number with 6 decimals"), "OrangeCenter")
    s7.add("J1", _("Time"), "OrangeCenter")
    s7.add("K1", _("Boolean"), "OrangeCenter")
    for row, color in enumerate(doc.colors.arr):
        s7.add(Coord("A2").addRow(row), color.name, color.name + "Left")
        s7.add(Coord("B2").addRow(row), 1,  color.name +"Integer")
        s7.add(Coord("C2").addRow(row), 1, color.name + "Integer")
        s7.add(Coord("D2").addRow(row), 1, color.name+ "Integer")
        s7.add(Coord("E2").addRow(row), 1, color.name+ "Integer")
        s7.add(Coord("F2").addRow(row), 1, color.name+ "Integer")
        s7.add(Coord("G2").addRow(row), 1, color.name+ "Integer")
        s7.add(Coord("H2").addRow(row), 1, color.name+ "Integer")
        s7.add(Coord("I2").addRow(row), 1, color.name+ "Integer")
        s7.add(Coord("J2").addRow(row), 1, color.name+ "Integer")
        s7.add(Coord("K2").addRow(row), 1, color.name+ "Integer")
    s7.addTotalsHorizontal("A11", ["Total", "#SUM", "#SUM", "#SUM", "#AVG","#MEDIAN","#SUM", "#AVG","#MEDIAN", "#SUM", "#SUM"])
    s7.addTotalsVertical("L1", ["Total", "#SUM", "#AVG","#MEDIAN","#SUM", "#AVG","#MEDIAN", "#SUM", "#SUM", "#SUM", "#SUM"])
    s7.freezeAndSelect("B2") 
    

    doc.setActiveSheet(s6)
    doc.save()

    return "demo_ods took {}".format(datetime.now()-start)
   
def demo_odt_commands(doc):
    doc.setMetadata("OfficeGenerator title",  "OfficeGenerator subject", "Turulomio")
    doc.setMetadata(_("OfficeGenerator ODT example"),  _("Demo with ODT documents"), "Turulomio", _("This file have been generated with OfficeGenerator-{}. You can see OfficeGenerator main page in http://github.com/turulomio/officegenerator").format(__version__), "officegenerator demo files")

    doc.title(_("Manual of officegenerator"))
    doc.subtitle(_("Version {}".format(__version__)))

    doc.header(_("ODT"), 1)
    doc.simpleParagraph(_("ODT files can be quickly generated with OfficeGenerator.") + " " + 
                                       _("It create predefined styles that allows to create nice documents without worry about styles."))

    doc.header(_("OfficeGenerator predefined paragraph styles"), 2)
    doc.simpleParagraph(_("OfficeGenerator has headers and titles as you can see in the document structure.") + " " + 
                                       _("Morever, it has the following predefined styles:"))
    doc.simpleParagraph(_("This is the 'Standard' style"))
    doc.simpleParagraph(_("This is the 'StandardCenter' style"), style='StandardCenter')
    doc.simpleParagraph(_("This is the 'StandardRight' style"), style='StandardRight')
    doc.simpleParagraph(_("This is the 'Illustration' style"), style='Illustration')
    doc.simpleParagraph(_("This is the 'Bold18Center' style"), style='Bold18Center')
    doc.simpleParagraph(_("This is the 'Bold16Center' style"), style='Bold16Center')
    doc.simpleParagraph(_("This is the 'Bold14Center' style"), style='Bold14Center')
    doc.simpleParagraph(_("This is the 'Bold12Center' style"), style='Bold12Center')
    doc.simpleParagraph(_("This is the 'Bold12Underline' style"), style='Bold12Underline')
    doc.pageBreak()

    doc.header(_("Tables"), 2)
    doc.simpleParagraph(_("We can create tables too, for example with size 11pt:"))
    doc.table(  [_("Concept"), _("Value")], 
                        [
                            [_("Text"), _("This is a text")], 
                            [_("Datetime"), datetime.now()], 
                            [_("Date"), date.today()], 
                            [_("Decimal"), Decimal("12.121")], 
                            [_("Currency"), Currency(12.12, "EUR")], 
                            [_("Percentage"), Percentage(1, 3)], 
                        ],
                        [3, 4],
                        11,
                        name=_("First"),
                    )
    doc.simpleParagraph(_("Tables with the size 10pt:"))
    doc.table(  [_("Concept"), _("Value")], 
                        [
                            [_("Text"), _("This is a text")], 
                            [_("Decimal"), Decimal("-12.121")], 
                            [_("Currency"), Currency(-12.12, "EUR")], 
                            [_("Percentage"), Percentage(-1, 3)], 
                        ], 
                        [4, 5], 
                        10
                    )
    doc.simpleParagraph(_("Tables with 8pt size:"))
    doc.table(  [_("Concept"), _("Value")], 
                        [
                            [_("Text"), _("This is a text")], 
                        ], 
                        [4,12], 
                        8,
                        name=_("Third")
                    )
    doc.pageBreak()

    doc.header(_("Lists and numbered lists"), 2) 
    
    doc.simpleParagraph(_("Simple list"))
    doc.list(   [   ["Prueba hola no. Prueba hola no. Prueba hola no. Prueba hola no. Prueba hola no. Prueba hola no. Prueba hola no. ", ], 
                        ["Adios", ], 
                        ["Bienvenido", ]
                    ],  list_style="List_20_1", paragraph_style="Standard")       
    doc.simpleParagraph(_("Multilevel list with bullets"))
    doc.list(   [   ["1", ["1.1", "1.2"]], 
                        ["2"], 
                        ["3",  ["3.1", ["3.1.1", "3.1.2"]]]
                    ],  list_style="List_20_1",  paragraph_style="Standard")   
    doc.simpleParagraph(_("Multilevel list with Numbers"))
    doc.list(   [   ["1", ["1.1", "1.2"]], 
                        ["2"], 
                        ["3",  ["3.1", ["3.1.1", "3.1.2"]]]
                    ],  list_style="Numbering_20_123", paragraph_style="Text_20_body")   
    doc.pageBreak()

    doc.header(_("Images"), 2)
    pngfile = pkg_resources.resource_filename(__name__, 'images/crown.png')#Gets package filename
    doc.addImage(pngfile,"images/crown.png")#Add image to the doc, adding a key "images/crown.png"
    p = P(stylename="Standard")
    p.addText("Este es un ejemplo de imagen as char: ")
    p.addElement(doc.image("images/crown.png", "3cm", "3cm"))
    p.addText(". Ahora sigo escribiendo sin problemas.")
    doc.insertInCursor(p, after=True)

    doc.simpleParagraph(_("As you can see, I can reuse it one hundred times. File size will not be increased because I used reference names."))
    p=P(stylename="Standard")
    for i in range(100):
        p.addElement(doc.image("images/crown.png", 0.2, 0.2, name="Crown.{}".format(i)))
    doc.insertInCursor(p, after=True)

    doc.simpleParagraph(_("The next paragraph is generated with the illustration method"))
    doc.illustration(["images/crown.png"]*5, 2.5,2.5, "IllustrationCrowns")

    doc.pageBreak(horizontal=True)
    doc.header(_("Horizontal page"), 2)
    doc.simpleParagraph(_("By default OfficeGenerator uses a predefined A4 page.") + " " +  _("Morever, you can set a predefined A4 horizontal page, as you can see."))
    doc.pageBreak()

    doc.header("ODS Writing", 1)
    doc.simpleParagraph("This library create several default styles for writing ODS files. You can see examples in officegenerator.ods.")
    doc.pageBreak()
    
    doc.header("Search and replace", 2)
    doc.simpleParagraph("You can search strings in a document and replace them programatically.")
    doc.simpleParagraph("__REPLACEME__")
    doc.simpleParagraph("You can delete this strings if you aren't going to use it")
    doc.simpleParagraph("__DELETEME__")
    
    doc.header("ODS Reading", 1)
    doc.pageBreak()

    doc.header("XLSX", 1)
    
def demo_odt_search_and_replace():
    start=datetime.now()
    doc=ODT("officegenerator_search_and_replace.odt", template="officegenerator.odt")
    doc.search_and_replace("_DELETEME__", "")
    doc.search_and_replace("__REPLACEME__",  "This text has been replaced programatically")
    doc.save()
    return "demo_odt_search_and_replace took {}".format(datetime.now()-start)

    
def demo_odt_standard():
    start=datetime.now()
    doc=ODT_Standard("officegenerator.odt")
    demo_odt_commands(doc)
    doc.save()
    return "demo_odt_standard took {}".format(datetime.now()-start)
    
def demo_odt_manual_styles():
    start=datetime.now()
    doc=ODT_Manual_Styles("officegenerator_manual_styles.odt")
    demo_odt_commands(doc)
    doc.save()
    return "demo_odt_manual_styles took {}".format(datetime.now()-start)

def demo_xlsx():
    start=datetime.now()
    xlsx=XLSX_Write("officegenerator.xlsx")
    xlsx.setCurrentSheet(0)

    xlsx.setSheetName(_("Styles"))
    xlsx.setColumnsWidth([20, 20, 20, 20, 20, 20, 20, 20])
    
    xlsx.overwrite("A1", _("Style name"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("B1", _("Date and time"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("C1", _("Date"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("D1", _("Integer"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("E1", _("Euros"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("F1", _("Percentage"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("G1", _("Number with 2 decimals"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("H1", _("Number with 6 decimals"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("I1", _("Time"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("J1", _("Boolean"), style=xlsx.stOrange,  alignment="center")
    for row, style in enumerate([xlsx.stOrange, xlsx.stGreen, xlsx.stGrayLight, xlsx.stYellow, xlsx.stGrayDark, xlsx.stWhite, None]):
        xlsx.overwrite(Coord("A2").addRow(row), xlsx.styleName(style), style=style)
        xlsx.overwrite(Coord("B2").addRow(row), datetime.now(), style=style)
        xlsx.overwrite(Coord("C2").addRow(row), date.today(), style=style)
        xlsx.overwrite(Coord("D2").addRow(row), pow(-1, row)*-10000000, style=style)
        xlsx.overwrite(Coord("E2").addRow(row), Currency(pow(-1, row)*12.56, "EUR"), style=style)
        xlsx.overwrite(Coord("F2").addRow(row), Percentage(1, 3), style=style,  decimals=row+1)
        xlsx.overwrite(Coord("G2").addRow(row), pow(-1, row)*12.121212, style=style, decimals=2)
        xlsx.overwrite(Coord("H2").addRow(row), pow(-1, row)*-12.121212, style=style, decimals=6)
        xlsx.overwrite(Coord("I2").addRow(row), datetime.now().time(), style=style, decimals=6)
        xlsx.overwrite(Coord("J2").addRow(row), True, style=style, decimals=6)
    xlsx.setComment("B2", _("This is a comment"))
    
    ##To write a custom cell
    cell=xlsx.wb.active['B12']
    cell.font=openpyxl.styles.Font(name='Arial', size=16, bold=True, color='00FF0000')#Red
    cell.value=_("This is a custom cell")
    #Merge cells
    xlsx.overwrite_and_merge("A13:C14", _("This cell is going to be merged with B13 and C13"),style=xlsx.stOrange)
    xlsx.overwrite_and_merge("A18:G18", _("This cell is going to be merged and aligned"),style=xlsx.stGrayDark, alignment="right")

    xlsx.overwrite("A16", [[_("This are booleans"), False, True]], style=xlsx.stWhite, alignment='right')
    xlsx.overwrite("D16", _("These are formulas returning booleans"), style=xlsx.stGreen)

    xlsx.overwrite("A20",  [["Una fila"]*3], style=xlsx.stGrayDark)
    xlsx.overwrite_and_merge("E13:G13", _("This sheet max rows are {} and max columns {}").format(xlsx.rowNumber(xlsx.ws_current_id), xlsx.columnNumber(xlsx.ws_current_id)), style=xlsx.stYellow,  alignment="center")

    #Named cells
    xlsx.overwrite_and_merge("A23:B23", _("Cell B23 has a name 'Amount"), style=xlsx.stWhite)
    xlsx.overwrite("C23", 5, style=xlsx.stWhite)
    xlsx.setCellName("$C$23", "Amount")

    xlsx.overwrite_and_merge("A24:B24", _("Cell B24 has a name 'Price"), style=xlsx.stWhite)
    xlsx.overwrite("C24", Currency(10,'EUR'), style=xlsx.stWhite)
    xlsx.setCellName("$C$24", "Price")

    xlsx.overwrite_and_merge("A25:B25", _("Cell B25 has a product with names"), style=xlsx.stWhite)
    xlsx.overwrite_formula("C25", "=Amount*Price", "€", style=xlsx.stWhite, alignment='right')

    xlsx.freezeAndSelect("A9","B11", "A9")
   
    
    #To text split and cur position
    xlsx.createSheet("Freeze A1 None")
    for letter in "ABCDEFGHIJKLMNOPQ":
        for number in range(1, 100):
            xlsx.overwrite(letter + str(number), letter+str(number), style=xlsx.stYellow)
    xlsx.freezeAndSelect("A1","Z199")
    #To text split and cur position
    xlsx.createSheet("Freeze A3 None")
    for letter in "ABCDEFGHIJKLMNOPQ":
        for number in range(1, 100):
            xlsx.overwrite(letter + str(number), letter+str(number), style=xlsx.stYellow)
    xlsx.freezeAndSelect("A3","Z199")
    #To text split and cur position
    xlsx.createSheet("Freeze C1 None")
    for letter in "ABCDEFGHIJKLMNOPQ":
        for number in range(1, 100):
            xlsx.overwrite(letter + str(number), letter+str(number), style=xlsx.stYellow)
    xlsx.freezeAndSelect("C1","Z199")
    #To text split and cur position
    xlsx.createSheet("Freeze C3 None")
    for letter in "ABCDEFGHIJKLMNOPQ":
        for number in range(1, 100):
            xlsx.overwrite(letter + str(number), letter+str(number), style=xlsx.stYellow)
    xlsx.freezeAndSelect("C3","Z199")
    
    xlsx.createSheet("Add totals example")
    xlsx.overwrite("A1", _("Style name"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("B1", _("Date and time"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("C1", _("Date"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("D1", _("Integer"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("E1", _("Euros"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("F1", _("Percentage"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("G1", _("Number with 2 decimals"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("H1", _("Number with 6 decimals"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("I1", _("Time"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("J1", _("Boolean"), style=xlsx.stOrange,  alignment="center")
    xlsx.overwrite("K1", _("Boolean"), style=xlsx.stOrange,  alignment="center")
    for row, style in enumerate([xlsx.stOrange, xlsx.stGreen, xlsx.stGrayLight, xlsx.stYellow, xlsx.stGrayDark, xlsx.stWhite, None]):
        xlsx.overwrite(Coord("A2").addRow(row), xlsx.styleName(style), style=style)
        xlsx.overwrite(Coord("B2").addRow(row), 1, style=style)
        xlsx.overwrite(Coord("C2").addRow(row), 1, style=style)
        xlsx.overwrite(Coord("D2").addRow(row), 1, style=style)
        xlsx.overwrite(Coord("E2").addRow(row), 1, style=style)
        xlsx.overwrite(Coord("F2").addRow(row), 1, style=style,  decimals=row+1)
        xlsx.overwrite(Coord("G2").addRow(row), 1, style=style, decimals=2)
        xlsx.overwrite(Coord("H2").addRow(row), 1, style=style, decimals=6)
        xlsx.overwrite(Coord("I2").addRow(row), 1, style=style, decimals=6)
        xlsx.overwrite(Coord("J2").addRow(row), 1, style=style, decimals=6)
        xlsx.overwrite(Coord("K2").addRow(row), 1, style=style, decimals=6)
    xlsx.overwriteTotalsHorizontal("A9", ["Total", "#SUM", "#SUM", "#SUM", "#AVG","#MEDIAN","#SUM", "#AVG","#MEDIAN", "#SUM", "#SUM"])
    xlsx.overwriteTotalsVertical("L1", ["Total", "#SUM", "#AVG","#MEDIAN","#SUM", "#AVG","#MEDIAN", "#SUM","#SUM"])
    xlsx.setColorScale("L2:L8")
    xlsx.freezeAndSelect("B2") 

    xlsx.save()
    return "demo_xlsx took {}".format(datetime.now()-start)

def demo_xlsx_updated():
    start=datetime.now()
    xlsx=XLSX_Write("officegenerator_updated.xlsx", "officegenerator.xlsx")
    xlsx.setCurrentSheet(0)
    
    xlsx.overwrite("A2", _("Orange"))
    xlsx.overwrite("A5", _("Yellow"))
    xlsx.overwrite("A7", _("White"), style=xlsx.stWhite, alignment="center")
    xlsx.overwrite("D4", 1223)
    #Merge cells
    xlsx.overwrite_and_merge("A15:C16", _("This cell is going to be merged with B13 and C13"),style=xlsx.stOrange)
    xlsx.overwrite_and_merge("A17:G17", _("This cell is going to be merged and aligned"),style=xlsx.stGrayDark, alignment="right")

    xlsx.save()
    return "demo_xlsx_updated took {}".format(datetime.now()-start)
    
def demo_xlsx_readonly(): 
    start=datetime.now()
    doc=XLSX_Read("officegenerator.xlsx")
    output=open("officegenerator_xlsx_readonly.txt", "w")
        
    range_=Range("A2:J2")
    for coord in range_.coords()[0]:
        output.write("{} {}\n".format(coord,  doc.getCellValue(0, coord)))
        
    output.write("{}\n".format(doc.getColumnValues(0, "J", skip_up=0)))
    output.write("{}\n".format(doc.getRowValues(1, "100", skip_left=3)))
        
    output.write("{}\n".format(doc.values_by_range(0, range_) ))
    output.close()
    return "demo_xlsx_readonly took {}".format(datetime.now()-start)
        
    
def demo_xlsx_data_only():
    start=datetime.now()   
    output=open("officegenerator_xlsx.data_only.txt", "w")

    # To read a XLSX_Write created file with data_only fails Values are not generated by libreoffice, just oopenpyxl
    doc=XLSX_Read("officegenerator.xlsx", data_only=True)
    output.write(str(doc.getCellValue(0, "C25"))+"\n")
    #File with formulas. Works. Openpyxl generated formulas
    doc=XLSX_Read("officegenerator.xlsx")
    output.write(str(doc.getCellValue(0, "C25"))+"\n")
    #File converted to data_only. Works to read value due libreoffice updated values in create_data_only_filename function
    doc=XLSX_Read("officegenerator.xlsx.rewritten.xlsx")
    output.write(str(doc.getCellValue(0, "C25"))+"\n")
    
    output.close()
    return "demo_xlsx_data_only took {}".format(datetime.now()-start)
    
def demo_standard_sheets():
    start=datetime.now()
    ods=ODS_Write("officegenerator_standard_sheets.ods")
    odt=ODT_Standard("officegenerator_standard_sheets.odt")
    xlsx=XLSX_Write("officegenerator_standard_sheets.xlsx")
    
    m=Model()
    m.setTitle("HV")
    m.setHorizontalHeaders(["Number", "Data", "More data", "Dt"], [1, 2, 3, 4])
    m.setVerticalHeaders(["V1", "V2", "V3"]*10, 4)
    data=[]        
    for row in range(30):
        data.append([row, "Data", "Data++", datetime.now()+timedelta(days=row)])
    m.setData(data)
    m.ods_sheet(ods)
    m.xlsx_sheet(xlsx)
    m.odt_table(odt, 15, 8)
    
    m2=Model()
    m2.setTitle("V")
    m2.setHorizontalHeaders(None, [1, 2, 3])
    m2.setVerticalHeaders(["Number", "Data", "More data"]*10)
    m2.setData(data)
    m2.removeColumns([1, 2, ])
    m2.removeRows([1, 2, ])
    m2.ods_sheet(ods)
    m2.xlsx_sheet(xlsx)
    m2.odt_table(odt, 15, 10)
    
    m=Model()
    m.setTitle("H totals")
    m.setHorizontalHeaders(["Concept", "Decimal", "Currency", "Percentage"], [5, 3, 3, 3])
    data=[]        
    for row in range(30):
        data.append([f"Concept {row}", row*10, Currency(row*10/7, "EUR"), Percentage(row, 12) ])
    m.setData(data)
    m.setHorizontalTotalDefinition(["Total", "#SUM","#AVG","#MEDIAN" ])
    m.ods_sheet(ods)
    m.xlsx_sheet(xlsx)
    m.odt_table(odt, 15, 8)    
    m=Model()

    m.setTitle("H totals skip 2")
    m.setHorizontalHeaders(["Concept", "Decimal", "Currency", "Percentage"], [5, 3, 3, 3])
    data=[]        
    for row in range(30):
        data.append([f"Concept {row}", row*10, Currency(row*10/7, "EUR"), Percentage(row, 12) ])
    m.setData(data)
    m.setHorizontalTotalDefinition(["Total", "#SUM","#AVG","#MEDIAN" ],totals_index_from=2)
    m.ods_sheet(ods)
    m.xlsx_sheet(xlsx)
    m.odt_table(odt, 15, 8)
    
    m=Model()
    m.setTitle("V totals")
    m.setHorizontalHeaders(["Concept", "Decimal", "Decimal2", "Decimal3"], [5, 3, 3, 3])
    data=[]        
    for row in range(30):
        data.append([f"Concept {row}", row*10, row*10, row*10 ])
    m.setData(data)
    m.setVerticalTotalDefinition(["Total"]+["#SUM"]*m.numDataRows() )
    m.ods_sheet(ods)
    m.xlsx_sheet(xlsx)
    m.odt_table(odt, 15, 8)    
    m=Model()

    m.setTitle("HV totals")
    hh=["Concept", "Decimal", "Decimal2", "Decimal3"]
    m.setHorizontalHeaders(hh, [5, 3, 3, 3])
    data=[]        
    for row in range(30):
        data.append([f"Concept {row}", row*10, row*10, row*10 ])
    m.setData(data)
    m.setHorizontalTotalDefinition(["Total", "#SUM","#AVG","#MEDIAN" ])
    m.setVerticalTotalDefinition(["Total"]+["#SUM"]*(m.numDataRows() +1))
    m.ods_sheet(ods)
    m.xlsx_sheet(xlsx)
    m.odt_table(odt, 15, 8)
    
    #Model auto
    hh=["Concept", "Decimal", "Decimal2", "Decimal3"]
    data=[]        
    for row in range(30):
        data.append([f"Concept {row}", row*10, -row*10, row*10 ])
    m=Model_Auto("Model_Auto", hh, data)
    m.order_with_none(1, reverse=True)
    m.ods_sheet(ods)
    m.xlsx_sheet(xlsx)
    m.odt_table(odt, 15, 8)
    
    
    #Model auto
    hh=["Concept", "Concept2", "Decimal", "Decimal2", "Decimal3"]
    data=[]        
    for row in range(30):
        data.append([f"Concept {row}", datetime.now(),  Currency(row*10, 'EUR'), Currency(-row*10, 'EUR'), Currency(row*10, 'EUR') ])
    m=Model_Auto("Model_Auto skip 2", hh, data, column_index_from=2)
    m.ods_sheet(ods)
    m.xlsx_sheet(xlsx)
    m.odt_table(odt, 15, 8)
    
    ods.save()
    xlsx.save()
    odt.save()
    
    m.ods_file("officegenerator_standard_sheets_direct.ods")
    m.xlsx_file("officegenerator_standard_sheets_direct.xlsx")
    return "demo_standard_sheets took {}".format(datetime.now()-start)


if __name__ == "__main__":
    main()
